;
; program NTSC5
;
; purpose: generate NTSC frames which are modifiable
;          by reading an input device
;
; NOTE: the timing of the vertical blanking scan line loop
;       is inconsistent; this may result in slight image
;       flickers during input
;
; IMPORTANT: all of these timings are specific to a 
;            CPU clock frequency of 3.6864 MHz
;
JP start

; data region

numColours:    EQU 9

; the purpose of these two is to provide a convenient way
; to reference this array via 8-bit pointers without
; generating assembler warnings
coloursLowByte: EQU 00h
coloursAfterLowByte: EQU coloursLowByte + numColours

colours:       EQU 0E0h << 8 + coloursLowByte
coloursAfter:  EQU 0E0h << 8 + coloursAfterLowByte

positionPtr:   EQU 0F000h

          ; pointer to where the white colour is found
		  
lastState:     EQU 0F010h  ; bit 0: button state
currentState:  EQU 0F020h

start:

; initialize variables
LD HL, colours
LD (positionPtr), HL ; we start at first position

; initial colours for each position
; they're all black, except for one white
LD A, 3              ; white
LD (colours), A
LD A, 1              ; black
LD (colours+1), A
LD (colours+2), A
LD (colours+3), A
LD (colours+4), A
LD (colours+5), A
LD (colours+6), A
LD (colours+7), A
LD (colours+8), A

; states
LD A, 0
LD (lastState), A
LD (currentState), A ; all states store "not pressed"

; setup registers to enter loop
LD C, 06h            ; output port
LD A, 1
LD E, 242

OUT (C), A           ; start with front porch level of 0.3V
NOP
NOP
NOP                  ; a small initial front porch

;
; visible scan line loop: generate 242 visible vertical scan lines
;
new_scanline:
; here, output is 0.3V
; here, register E contains remaining visible scan line count
; here, register C contains output port number
; here, register A contains 1
; (since logical 1 outputs 0.3V: front porch level)
                     ; here we're on the front porch
                     
DEC A                ; 4 clocks
                     ; A := 0 (preparing to output 0V for horizontal sync)
OUT (C), A           ; 12 clocks
                     ; falling edge to 0V (logical 0)
                     
                     ; we're now at the start of horizontal sync

INC A                ; 4 clocks, still at 0V
                     ; A := 1, preparing to output 0.3V
OUT (C), A           ; 12 clocks, rising edge to 0.3V (logical 1)
                     ; signal has now risen to the level of the
                     ; back porch
                     
					 ; TOTAL TIME ON HORIZONTAL SYNC
                     ;                   4 + 12 = 16 clocks
                     
                     ; we're at the start of the back porch
                     ; and have risen to 0.3V (logical 1)                  
                     

; front porch took: 30 clocks
; back porch took: 19 clocks
; horizontal sync took: 16 clocks
; total invisible: 65 clocks
;
; 1 clock = 1 / 3.6864 MHz = 0.271267361us
; total scan line: 63.55us / 0.271267361us = 234.27 clocks per line
;                                            round down to 234
; remaining for visible region: 234 - 65 = 169 clocks
; NOTE: visible region starts on back porch

; IMPORTANT:
;     output voltage is now 0.3V from the back porch
;     available visible region MUST NOT change that output for
;     another 19 clocks

                     ; we're still on back porch

; BEGIN 169 clock area

LD A, R              ; 9 clocks, NOOP instruction

LD HL, colours       ; 16 clocks
                     ; HL := pointer to colour array
					 
OUTI                 ; 16 clocks
                     ; output byte at address in HL to port C
                     ; B := B - 1
					 ; HL := HL + 1
					 ; this changes colour being output and
					 ; advances colour pointer

					 ; NOTE: the back porch can be longer, if
					 ;       available visible area begins
					 ;       with black
					 ;       this 19 clock internval represents
					 ;       the minimum time needed for back porch

					 ; we're now on the available visible region

OUTI                 ; 16 clocks
OUTI                 ; 16 clocks
OUTI                 ; 16 clocks
OUTI                 ; 16 clocks
OUTI                 ; 16 clocks
OUTI                 ; 16 clocks
OUTI                 ; 16 clocks
OUTI                 ; 16 clocks

; END 169 clock area
					 
; end of available visible region

; here, register C contains output port number
                     ; we're still in visible region
                     ; assume register C contains output port number

LD A, 1              ; 7 clocks, still in visible region
OUT (C), A           ; 12 clocks, signal 0.3V (logical 1)
                     
                     ; 7 + 12 = 19 clocks taken out of visible region
                     ; to set up front porch voltage

                     ; we're now at the start of the front porch

					 ; 14 clocks (from jumping to start of loop)
					 ; plus
                     ; 4 + 12 = 16 clocks (from setting up
                     ; horizontal sync at beginning of loop)
					 
					 ; TOTAL TIME ON FRONT PORCH
                     ;                14 + 16 = 30 clocks
					 
DEC E                ; 4 clocks
JP NZ, new_scanline  ; 10 clocks
                     ; E-- and next visible scan line
					 ; this spends an additional 14 clocks on front porch

					 
; we've now completed all visible scan lines

; here, register C contains output port number
; here, register A contains 1
; (since logical 1 outputs 0.3V: front porch level)

LD E, 2              ; 7 clocks
                     ; setup vertical blanking line counter
					 ; NOTE: there might be great variability
					 ;       in televisions here; mine worked
					 ;       better with a low value, rather
					 ;       than the value of 20 from specification

DEC A                ; 4 clocks
                     ; A := 0 (preparing to output 0V vblank)
OUT (C), A           ; 12 clocks
                     ; falling edge to 0V (logical 0)
					 
					 ; we're now at the start of vblank
					 ; (right after horizontal sync)
;
; vertical blanking loop: generate several invisible vertical
;                         sync lines (inverted horizontal sync)
;
new_vblank_line:
; here, register E contains remaining vertical blanking line count
; here, register C contains output port number

                     ; we're now at the start of the vblank portion
					 ; outputting 0V
					 
; 19 clocks to set up horizontal sync
; 16 clocks spent on horizontal sync
; 27 clocks pre-empting values in case we are jumping to visible
;           lines loop
; 16 clocks spent counter decrement/jump to start of loop
; 19 + 16 + 11 + 16 = 62 total clocks for housekeeping
;
; 234 clocks per line
; 234 - 62 = 172 clocks left for computation
; 

; BEGIN 172 clock area

; old state := current state
LD HL, lastState     ; 7 clocks
LD A, (currentState) ; 13 clocks
LD (HL), A           ; 13 clocks

; read new state
IN A, (03h)          ; 11 clocks
                     ; bit 0: current button state
LD (currentState), A ; 13 clocks

; check button's current state
BIT 0, A             ; 8 clocks
                     ; is it pressed now?
JP NZ, endChecks     ; 10 clocks
; it's pressed now
BIT 0, (HL)          ; 12 clocks
                     ; was it pressed last time?
JP Z, endChecks      ; 10 clocks
; no, so we act on a button press

LD HL, (positionPtr) ; 16 clocks
                     ; HL := ptr to current position
LD (HL), 1           ; 10 clocks
                     ; set current position to black
INC L                ; 4 clocks
                     ; current position++
LD A, coloursAfterLowByte   ; 7 clocks
                     ; A := ptr to past array end
CP L                 ; 4 clocks
JP NZ, afterChecks   ; 10 clocks
                     ; we haven't wrapped around
; we wrapped around

LD L, coloursLowByte ; 7 clocks
                     ; HL := ptr to first position
					 
afterChecks:
LD (HL), 3           ; 10 clocks
                     ; make current position colour white
LD (positionPtr), HL ; 16 clocks
                     ; store current position

endChecks:

; END 172 clock area

LD A, 1              ; 7 clocks, still in vblank portion
OUT (C), A           ; 12 clocks, signal 0.3V (logical 1)
                     ; 7 + 12 = 19 clocks spent in vblank
					 
					 ; we're now at the start of horizontal sync
					 
					 ; it's inverted for vertical blanking
					 ; interval (0.3V instead of 0V)

DEC A                ; 4 clocks
                     ; A := 0
OUT (C), A           ; 12 clocks, signal 0V (logical 0)
                     ; 4 + 12 = 16 clocks spent on horizontal sync
					 
					 ; we're now at the start of vblank portion

LD B, 242            ; 7 clocks
INC A                ; 4 clocks
                     ; A := 1, B := 242, HL := ptr to colour data
                     ; pre-empt the values needed by visible
					 ; scan line loop, in case we're
					 ; done with the vertical blanking
					 ; 7 + 4 = 11 clocks
					 ; NOTE: this IS wasteful, but is done this
					 ;       way to keep timing consistent
					 ;       and not have to unroll loops

DEC E                ; 4 clocks
                     ; vertical blanking line counter--
JR NZ, new_vblank_line
                     ; 12 clocks when jumping
					 ; 7 clocks when not

; now return to visible scan line loop to begin a new video frame
					 
; here, register B contains remaining visible scan line count
; here, register C contains output port number
; here, register A contains 1

OUT (C), A           ; 12 clocks, signal 0.3V (logical 1)
                     ; we're now on the front porch of first
					 ; visible scan line

LD E, B              ; 4 clocks
                     ; prepare visible line loop

JP new_scanline      ; 10 clocks
